#!/usr/bin/python
# -*- coding:utf-8 -*-
"""
Migrate elbv1 resource to elbv2

Step1: check elbv1 resource can be migrated or not
Step2: create resource in elbv2
Setp3: delete the resource in elbv1 if necessary

Usage:
    migrate.py
    [--lb] <value>
    [--dry-run]

eg: 
    python migrate.py
    python migrate.py --lb xxxx
    python migrate.py --lb 'xxxx xxxx'
    python migrate.py --dry-run
"""

import argparse
import ConfigParser
import json
import logging
import math
import os
import requests
import signal
import sys
import time

from functools import wraps
from requests.packages.urllib3.exceptions import InsecureRequestWarning


requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
logging.getLogger("requests").setLevel(logging.WARNING)


"""set log"""
logger = logging.getLogger('running_log')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
file_handler = logging.FileHandler('logger.log')
file_handler.setFormatter(formatter)
file_handler.setLevel(logging.DEBUG)

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.INFO)

logger.addHandler(file_handler)
logger.addHandler(stream_handler)


"""result log"""
result_logger = logging.getLogger('result_log')
result_logger.setLevel(logging.INFO)
result_file_handler = logging.FileHandler('output', 'w')
result_stream_handler = logging.StreamHandler()
result_file_handler.setLevel(logging.INFO)
result_stream_handler.setLevel(logging.INFO)
result_logger.addHandler(result_file_handler)
result_logger.addHandler(result_stream_handler)


"""check user.cfg is existed or not"""
if not os.path.isfile("user.cfg"):
    result_logger.error('can not find configuration file: user.cfg, please config user.cfg')
    sys.exit(1)

"""get config from run.cfg"""
cf = ConfigParser.ConfigParser()
cf.read("user.cfg")
USERNAME = cf.get('env', 'username')
PASSWORD = cf.get('env', 'password')
TENANT_ID = cf.get('env', 'project_id')
DOMAIN_NAME = cf.get('env', 'domain_name')
IAM_ENDPOINT = cf.get('env', 'iam_endpoint')
VPC_ENDPOINT = cf.get('env', 'vpc_endpoint')
AS_ENDPOINT = cf.get('env', 'autoscaling_endpoint')
ELBV1_ENDPOINT = cf.get('env', 'elb_endpoint')


API_RETRY_TIMES = 3
TIMEOUT = 20
MESSAGE_LENGTH = 48

"""iam API"""
IAM_API  = "/v3/auth/tokens"


"""AS API"""
AS_API = "/autoscaling-api/v1/%s/scaling_group"


"""quota API"""
QUOTA_API = "/v1/%s/quotas"


"""FloatingIP API"""
FIPS_API = "/v2.0/floatingips"
FIP_API = "/v2.0/floatingips/%s"

"""elbv1 API"""
LBS_API = "/v1.0/%s/elbaas/loadbalancers"
LB_API = "/v1.0/%s/elbaas/loadbalancers/%s"
KEEP_EIP_DELETE_API = "/v1.0/%s/elbaas/loadbalancers/%s/keep-eip"
LISTENERS_API = "/v1.0/%s/elbaas/listeners"
LISTENER_API = "/v1.0/%s/elbaas/listeners/%s"
LB_LISTENER_API = "/v1.0/%s/elbaas/listeners?loadbalancer_id=%s"
HEALTHCHECK_API = "/v1.0/%s/elbaas/healthcheck/%s"
MEMBERS_API = "/v1.0/%s/elbaas/listeners/%s/members"
MEMBER_API = "/v1.0/%s/elbaas/listeners/%s/members/action"
CERT_API = "/v1.0/%s/elbaas/certificate?id=%s"
JOB_API = "/v1.0/%s/jobs/%s"


"""elbv2 API"""
ELBV2_LBS_API = "/v2.0/lbaas/loadbalancers"
ELBV2_LB_API = "/v2.0/lbaas/loadbalancers/%s"
ELBV2_LISTENERS_API = "/v2.0/lbaas/listeners"
ELBV2_LISTENER_API = "/v2.0/lbaas/listeners/%s"
ELBV2_POOLS_API = "/v2.0/lbaas/pools"
ELBV2_POOL_API = "/v2.0/lbaas/pools/%s"
ELBV2_MEMBERS_API = "/v2.0/lbaas/pools/%s/members"
ELBV2_MEMBER_API = "/v2.0/lbaas/pools/%s/members/%s"
ELBV2_HEALTHMONITORS_API = "/v2.0/lbaas/healthmonitors"
ELBV2_HEALTHMONITOR_API = "/v2.0/lbaas/healthmonitors/%s"
ELBV2_CERTIFICATES_API = "/v2.0/lbaas/certificates"
ELBV2_CERTIFICATE_API = "/v2.0/lbaas/certificates/%s"
ELBV2_SUBNETS = "/v1/%s/subnets?vpc_id=%s"
ELBV2_PRIVATE_IPS = "/v1/%s/subnets/%s/privateips"
ELBV2_PORTS = "/v1/ports"
NEUTRON_SUBNET = "/v2.0/subnets/%s"


"""API valid status_code dict"""
VALID_API_CODE = {
    'get': [200],
    'post': [200, 201],
    'put': [200, 202],
    'delete': [204, 200]
}
AUTHORIZED_FAILED = 401
RESOURCE_NOT_FOUND = 404


"""IAM TOKEN"""
IAM_TOKEN = ''


def user_exit(signum, frame):
    print('\n')
    print('Bye Bye')
    sys.exit()


def get_token():
    post_data = {
        "auth": {
            "identity": {
                "methods": ["password"],
                "password": {
                    "user": {
                        "name": USERNAME,
                        "password": PASSWORD,
                        "domain": {
                            "name": DOMAIN_NAME
                        }
                    }
                }
            },
            "scope": {
                "project": {
                    "id": TENANT_ID
                }
            }
        }
    }
    post_data = json.dumps(post_data)
    global IAM_TOKEN
    IAM_TOKEN = process_request(IAM_API, 'iam', 'post', post_data)
    if not IAM_TOKEN:
        logger.error('get iam token failed')
        return False
    return IAM_TOKEN


def process_request(url, type='elbv1', action='get', body=None):
    """the base method for call api"""
    endpoint_dict = {
        'elbv1': ELBV1_ENDPOINT,
        'elbv2': VPC_ENDPOINT,
        'iam': IAM_ENDPOINT,
        'fip': VPC_ENDPOINT,
        'quota': VPC_ENDPOINT,
        'as': AS_ENDPOINT
    }
    endpoint = endpoint_dict[type]
    headers = {'content-type': 'application/json'}
    if type != 'iam':
        headers['X-Auth-Token'] = IAM_TOKEN if IAM_TOKEN else get_token()
        if not headers['X-Auth-Token']:
            return False
    url = 'https://' + endpoint + url
    logger.debug('begin to call %s: ' % url)
    retry = 0
    while 1:
        try:
            time.sleep(1)
            if action == 'post':
                res = requests.post(url, data=body, headers=headers, verify=False, timeout=TIMEOUT)
            if action == 'put':
                res = requests.put(url, data=body, headers=headers, verify=False, timeout=TIMEOUT)
            elif action == 'get':
                res = requests.get(url, headers=headers, verify=False, timeout=TIMEOUT)
            elif action == 'delete':
                res = requests.delete(url, headers=headers, verify=False, timeout=TIMEOUT)
            break
        except requests.exceptions.ConnectTimeout as exc:
            logger.error('Can not connect to endpoint: %s' % endpoint)
            retry += 1
        except requests.exceptions.Timeout as exc:
            logger.error('call api timeout, api: %s' % url)
            retry += 1
        except requests.exceptions.ConnectionError as exc:
            logger.error('can not connect to %s' % url)
            retry += 1
        except Exception as exc:
            logger.error('get response from %s failed' % url)
            retry += 1
        if retry >= API_RETRY_TIMES:
            logger.error('Have retried 3 times, stop calling %s' % url)
            return False
        logger.info('retry to process request to %s' % url)
    code = res.status_code
    if code == AUTHORIZED_FAILED:
        logger.error('authorized failed, please check your username/password/domain_name/project_id')
        return False
    elif code not in VALID_API_CODE[action]:
        logger.error('response status code %s is not valid' % code)
        logger.error('get response from %s failed' % url)
        logger.error('error message is: %s', res.text)
        return False
    logger.debug('call %s succeeded' % url)
    if type == 'iam':
        return res.headers['X-Subject-Token']
    if res.text:
        return json.loads(res.text)
    else:
        return True


def convert_ip_to_num(ip):
    ip_list = ip.split('.')
    num = 0
    i = 3
    for ip in ip_list:
        num += math.pow(256, i) * int(ip)
        i -= 1
    return int(num)


class check(object):

    def __init__(self, lbs):
        self.lbs = lbs
        self.certificates = []
        self.valid_lbs = []
        self.valid_lb_infos = []
        self.invalid_lbs = {}
        self.use_autoscaling_listener = []
        self.use_autoscaling_lbs = {}
        self.invalid_healthcheck_lbs = {}
        self.invalid_security_group_lbs = []
        self.warning_msg = []
        self.check_conf()
        self.get_as()
        self.check(lbs)

    def check_conf(self):
        """check user.cfg is empty or not"""
        logger.info('check configuration ...')
        if not USERNAME and not PASSWORD and not TENANT_ID and not DOMAIN_NAME:
            logger.error('please config username/password/project_id/domain_name in user.cfg')
        elif not IAM_ENDPOINT:
            logger.error('please config iam endpoint in user.cfg')
        elif not AS_ENDPOINT:
            logger.error('please config autoscaling endpoint in user.cfg')
        elif not ELBV1_ENDPOINT:
            logger.error('please config elb endpoint in user.cfg')
        elif not VPC_ENDPOINT:
            logger.error('please config vpc endpoint in user.cfg')
        else:
            iam_list = IAM_ENDPOINT.split('.')
            elbv1_list = ELBV1_ENDPOINT.split('.')
            vpc_list = VPC_ENDPOINT.split('.')
            as_list = AS_ENDPOINT.split('.')
            if not (iam_list[1] == elbv1_list[1] == vpc_list[1] == as_list[1]):
                logger.warn('iam and elbv1 and vpc are not belong to the same region: ')
                logger.warn('iam region: %s' % iam_list[1])
                logger.warn('autoscaling region: %s' % as_list[1])
                logger.warn('elb region: %s' % elbv1_list[1])
                logger.warn('vpc region: %s' % vpc_list[1])
            else:
                return
        sys.exit(1)

    def get_as(self):
        res = process_request(AS_API % TENANT_ID, 'as')
        if res == False:
            logger.error('can not get response from auto scaling api')
            sys.exit(1)
        for scaling_group in res['scaling_groups']:
            listener_id = scaling_group['lb_listener_id']
            if listener_id:
                self.use_autoscaling_listener.append(listener_id)

    def get_all_elbv1_lbs(self):
        res = process_request(LBS_API % TENANT_ID)
        if res == False:
            logger.error('get elbv1 loadbalancers failed')
            sys.exit(1)
        return res
    
    def get_elbv1_lb(self, lb_id):
        res = process_request(LB_API % (TENANT_ID, lb_id))
        if res == False:
            logger.error('get elbv1 loadbalancer failed, id is %s' % lb_id)
            sys.exit(1)
        return res

    def get_all_elbv1_listeners(self):
        res = process_request(LISTENERS_API % TENANT_ID)
        if res == False:
            logger.error('get elbv1 listeners failed')
            sys.exit(1)
        return res
    
    def get_elbv1_lb_listeners(self, lb_id):
        url = LB_LISTENER_API % (TENANT_ID, lb_id)
        res = process_request(url)
        if res == False:
            logger.error('get elbv1 listeners failed, loadbalancer is %s' % lb_id)
            sys.exit(1)
        return res
    
    def get_elbv1_healthcheck(self, _id):
        url = HEALTHCHECK_API % (TENANT_ID, _id)
        res = process_request(url)
        if res == False:
            logger.error('get elbv1 healthcheck failed, healthcheck id is %s' % _id)
            sys.exit(1)
        return res

    def get_elbv1_certificate(self, _id):
        url = CERT_API % (TENANT_ID, _id)
        res = process_request(url)
        if res == False:
            logger.error('get elbv1 certificate failed, certificate id is %s' % _id)
            sys.exit(1)
        return res

    def get_elbv1_members(self, listener_id):
        url = MEMBERS_API % (TENANT_ID, listener_id)
        res = process_request(url)
        if res == False:
            logger.error('get elbv1 member failed, listener id is %s' % listener_id)
            sys.exit(1)
        return res
 
    def check_rest_lb_quotas(self):
        res = process_request(QUOTA_API % TENANT_ID, 'elbv2')
        if res == False:
            logger.error('get elbv2 quota failed')
            sys.exit(1)
        res = res['quotas']['resources']
        for info in res:
            if info['type'] == 'loadbalancer':
                used_lb_num = info['used']
                lb_quota = info['quota']
            elif info['type'] == 'listener':
                used_listener_num = info['used']
                listener_quota = info['quota']
        rest_lb = lb_quota if lb_quota == -1 else lb_quota - used_lb_num
        rest_listener = listener_quota if listener_quota == -1 else listener_quota - used_listener_num
        if rest_lb < len(self.valid_lbs) and rest_lb != -1:
            logger.warn('Only can create %s loadbalancers at most, '
                         'please add loadbalancer quota' % rest_lb)
            return False
        if rest_listener < self.listener_num and rest_listener != -1:
            logger.warn('Only can create %s listeners at most, '
                         'please add listener quota' % rest_listener)
            return False
        return True

    def get_all_lbs(self, lb):
        """get all needed information"""
        lbs = []
        listeners = []
        if lb == 'all':
            lbs = self.get_all_elbv1_lbs()
            if lbs['loadbalancers']:
                lbs = lbs['loadbalancers']
            else:
                logger.warn('you have no elbv1 loadbalancers')
                sys.exit()
            listeners = self.get_all_elbv1_listeners()
        else:
            lb_ids = lb.split(' ')
            for lb_id in lb_ids:
                listener = self.get_elbv1_lb_listeners(lb_id)
                lb = self.get_elbv1_lb(lb_id)
                lbs.append(lb)
                listeners.extend(listener)
        for listener in listeners:
            healthcheck_id = listener['healthcheck_id']
            # get healthcheck detail
            if healthcheck_id:
                listener['healthcheck'] = self.get_elbv1_healthcheck(healthcheck_id)
            # get member detail
            members = self.get_elbv1_members(listener['id'])
            if members:
                listener['members'] = members
            certificate_id = listener['certificate_id']
            # get certificate detail
            if certificate_id:
                cert = self.get_elbv1_certificate(certificate_id)
                if len(cert['certificates']) == 0:
                    logger.error('get certificate error, certificate id is %s' % certificate_id)
                    return False
                listener['certificate'] = cert['certificates'][0]
                for cert in self.certificates:
                    if cert['id'] == listener['certificate']['id']:
                        break
                else:
                    self.certificates.append(listener['certificate'])
            for each_lb in lbs:
                if each_lb['id'] == listener['loadbalancer_id']:
                    if 'listeners' not in each_lb:
                        each_lb['listeners'] = [listener]
                    else:
                        each_lb['listeners'].append(listener)
        return lbs

    def set_invalid_listener_data(self, lb_id, listener_id, msg):
        msg = msg.ljust(MESSAGE_LENGTH, ' ')
        item = {
            'id': listener_id,
            'msg': msg
        }
        if lb_id in self.invalid_lbs and 'listeners' in self.invalid_lbs[lb_id]:
            for listener in self.invalid_lbs[lb_id]['listeners']:
                if listener_id == listener['id']:
                    break
            else:
                self.invalid_lbs[lb_id]['listeners'].append(item)
        else:
            self.invalid_lbs[lb_id] = {'listeners': [item]}

    def check(self, lb):
        """check lb can be migrated or not"""
        self.lbs = self.get_all_lbs(lb)
        self.lb_num = self.listener_num = self.listener_has_member = 0
        for res in self.lbs:
            valid_flag = True
            # admin_state_up can not be false
            logger.info('check lb admin_state_up')
            if res['admin_state_up'] not in [1, 'true']:
                msg = 'the loadbalancer is not activated'
                msg = msg.ljust(MESSAGE_LENGTH, ' ')
                self.invalid_lbs[res['id']] = {'msg': msg}
                valid_flag = False
                continue
            # check lb status
            logger.info('check lb status')
            if res['status'] != 'ACTIVE':
                msg = 'the status of loadbalancer is not active'
                msg = msg.ljust(MESSAGE_LENGTH, ' ')
                self.invalid_lbs[res['id']] = {'msg': msg}
                valid_flag = False
                continue
            # check internal lb security group
            # if lb has security group, it can still be migrated to elbv2
            # but we will give a warning to tell the user to set whitelist in elbv2
            logger.info('check lb security group')
            if res['type'] == 'Internal' and res['security_group_id']:
                msg = "elbv1 loadbalancer %s has setted security group, you must set elbv2 "\
                      "whitelist after migrated to elbv2" % res['id']
                logger.warn(msg)
                self.warning_msg.append(msg)
                self.invalid_security_group_lbs.append(res['id'])
            if 'listeners' in res:
                for listener in res['listeners']:
                    # check listener admin_state_up
                    logger.info('check listener admin_state_up')
                    if listener['admin_state_up'] not in [1, 'true']:
                        msg = 'the admin state up of listener is not right'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # check listener status
                    logger.info('check listener status')
                    if listener['status'] != 'ACTIVE':
                        msg = 'listener status is not active'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # check used scaling group or not
                    logger.info('check scaling group')
                    if listener['id'] in self.use_autoscaling_listener:
                        msg = 'listener has autoscaling group'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # udp/ssl protocol is not supported by elbv2
                    logger.info('check listener protocol')
                    if listener['protocol'] in ['UDP', 'SSL']:
                        msg = 'listener protocol is ' + listener['protocol']
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # ssl protocol can only be TLSv1.2 in elbv2
                    logger.info('check listener ssl protocol')
                    if listener['protocol'] == 'HTTPS' and listener['ssl_protocols'] == 'TLSv1.2':
                        msg = 'listener ssl protocol is TLSv1.2'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # check tcp session timeout
                    logger.info('check listener session sticky')
                    if listener['session_sticky']:
                        if listener['protocol'] == 'TCP' and listener['tcp_timeout'] > 1:
                            msg = 'listener tcp timeout is larger than 1'
                            self.set_invalid_listener_data(res['id'], listener['id'], msg)
                            valid_flag = False
                            continue
                        elif listener['protocol'] in ['HTTP', 'HTTPS'] and \
                            listener['cookie_timeout'] < 1440:
                            msg = 'listener http/https timeout is less than 1440'
                            self.set_invalid_listener_data(res['id'], listener['id'], msg)
                            valid_flag = False
                            continue
                    # elbv2 do not support sni
                    logger.info('check listener cert')
                    if listener['certificates']:
                        msg = 'listener certificate has sni'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # elbv2 do not support unhealthy threshold
                    logger.info('check listener healthcheck')
                    if 'healthcheck' in listener and \
                        listener['healthcheck']['unhealthy_threshold'] != listener['healthcheck']['healthy_threshold']:
                        msg = 'healthy and unhealthy threshold is not equal'
                        self.set_invalid_listener_data(res['id'], listener['id'], msg)
                        valid_flag = False
                        continue
                    # check health check status
                    logger.info('check member healthcheck status')
                    if 'members' in listener:
                        for member in listener['members']:
                            if member['health_status'] != 'NORMAL':
                                msg = 'listener member status is not normal'
                                self.set_invalid_listener_data(res['id'], listener['id'], msg)
                                valid_flag = False
                                continue
                    if valid_flag:
                        self.listener_num += 1
                        if listener['member_number']:
                            self.listener_has_member += 1
            if valid_flag:
                self.lb_num += 1
                self.valid_lb_infos.append(res)
                self.valid_lbs.append(res['id'])

    def common_print(self):
        result_logger.info('-' * 74)
        result_logger.info('| loadbalancer id' + ' ' * 19 + '| listener id' + ' ' * 24 + '|')
        result_logger.info('-' * 74)

    def print_result(self):
        result_logger.info('\n')
        result_logger.info('check result:')
        result_logger.info('*' * 40 + ' CHECK RESULT ' + '*' * 40)
        if self.valid_lbs:
            result_logger.info('loadbalancers below can be migrated to elbv2: ')
            result_logger.info('-' * 37)
            result_logger.info('| loadbalancer id                   |')
            result_logger.info('-' * 37)
            for lb_id in self.valid_lbs:
                result_logger.info( '| ' + lb_id + '  |')
                result_logger.info('-' * 37)
        if self.invalid_lbs:
            if self.valid_lbs:
                result_logger.info('\n')
            result_logger.info('loadbalancers below can not be migrated to elbv2: ')
            result_logger.info('-' * 124)
            result_logger.info('| loadbalancer id' + ' '  * 19 + '| listener id' + 
                               ' ' * 24 + '| invalid reason' + ' ' * 34 + '|')
            result_logger.info('-' * 124)
            for lb_id in self.invalid_lbs:
                if 'listeners' in self.invalid_lbs[lb_id]:
                    for listener in self.invalid_lbs[lb_id]['listeners']:
                        result_logger.info('| ' + lb_id + '  | ' + listener['id'] + 
                                           '   | ' + listener['msg'] + '|')
                        result_logger.info('-' * 124)
                else:
                    result_logger.info('| ' + lb_id + '  | ' + ' ' * 35 + 
                                       '| ' + self.invalid_lbs[lb_id]['msg'] + '|')
                    result_logger.info('-' * 124)
        else:
            result_logger.info('\n')
            for msg in self.warning_msg:
                result_logger.info('[Warning] ' + msg)
        result_logger.info('*' * 45 + ' END ' + '*' * 45)
        result_logger.info('\n')
        if not self.invalid_lbs:
            logger.info('check elbv2 quota...')
            if self.check_rest_lb_quotas():
                logger.info('Rest elbv2 quota is available')
                return True
            else:
                logger.error('Rest elbv2 quota is not available')
                return False
        return True
            

class migration(object):

    def __init__(self, lb_ids, lb_infos, certificates):
        self.lb_ids = lb_ids
        self.lb_infos = lb_infos
        self.certificates = certificates
        self.created_certificates = {}
        self.created_failed_certificates = []
        self.lb_num = len(lb_ids)
        self.created_mapping_resource = {}
        self.created_lbs = {}
        self.created_failed_lbs = []
        self.lb_algorithm_mappings = {
            'roundrobin': 'ROUND_ROBIN',
            'leastconn': 'LEAST_CONNECTIONS',
            'source': 'SOURCE_IP'
        }
        self.subnet_info = {}
        
    def before_migrate(self):
        pass

    def check_valid_vip(self, vpc_id):
        subnets_info = process_request(ELBV2_SUBNETS % (TENANT_ID, vpc_id), 'elbv2')
        if subnets_info == False:
            return False
        subnets_info = subnets_info['subnets']
        rest_num = num = 0
        for subnet in subnets_info:
            neutron_subnet_id = subnet['neutron_subnet_id']
            neutron_subnet_info = process_request(NEUTRON_SUBNET % neutron_subnet_id, 'elbv2')
            if neutron_subnet_info == False:
                result_logger.error('get neutron subnet failed, subnet id is %s' % neutron_subnet_id)
                return False
            allocation_pools = neutron_subnet_info['subnet']['allocation_pools']
            for each_pool in allocation_pools:
                start = convert_ip_to_num(each_pool['start'])
                end = convert_ip_to_num(each_pool['end'])
                num = num + end - start
                ip_pool = {
                    'start': start,
                    'end': end
                }
                if subnet['id'] in self.subnet_info:
                    self.subnet_info[neutron_subnet_id].append(ip_pool)
                else:
                    self.subnet_info[neutron_subnet_id] = [ip_pool]
            private_ips = process_request(ELBV2_PRIVATE_IPS % (TENANT_ID, subnet['id']), 'elbv2')
            if private_ips == False:
                result_logger.error('get private ips failed, subnet id is %s' % subnet['id'])
                return False
            used_num = len(private_ips['privateips'])
            if num - used_num > 0:
                rest_num = rest_num + num - used_num
                subnet_id = neutron_subnet_id
        if rest_num > len(self.lb_ids):
            return subnet_id
        else:
            return False
        
    def get_member_subnet_id(self, ip):
        ip_num = convert_ip_to_num(ip)
        for subnet_id in self.subnet_info:
            for ip_pool in self.subnet_info[subnet_id]:
                if ip_num >= ip_pool['start'] and \
                    ip_num <= ip_pool['end']:
                    return subnet_id
        return False
        
    def create_lb(self, subnet_id, lb_info):
        logger.info('begin to create elbv2 lb...')
        post_data = {
            "loadbalancer": {
                "name": lb_info['name'],
                "description": lb_info['description'],
                "vip_subnet_id": subnet_id
            }
        }
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_LBS_API, 'elbv2', 'post', post_data)
        if res == False:
            logger.error('create elbv2 loadbalancer failed')
            return False
        logger.info('create elbv2 loadbalancer succeeded')
        self.lb_num -= 1
        return res

    def delete_lb(self, lb_id):
        logger.info('begin to delete elbv2 lb, id is: %s' % lb_id)
        res = process_request(ELBV2_LB_API % lb_id, 'elbv2', 'delete')
        if res == False:
            logger.error('delete elbv2 loadbalancer failed, id is %s' % lb_id)
            return False
        logger.info('delete elbv2 loadbalancer succeeded')
        return res

    def create_listener(self, lb_id, listener_info):
        logger.info('begin to create elbv2 listener...')
        post_data = {
            'listener': {
                'protocol_port': listener_info['port'],
                'protocol': listener_info['protocol'],
                'description': listener_info['description'],
                'loadbalancer_id': lb_id,
                'name': listener_info['name']
            }
        }
        if listener_info['protocol'] == 'HTTPS':
            post_data['listener']['protocol'] = 'TERMINATED_HTTPS'
        if 'default_tls_container_ref' in listener_info:
            post_data['listener']['default_tls_container_ref'] = \
                listener_info['default_tls_container_ref']
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_LISTENERS_API, 'elbv2', 'post', post_data)
        if res == False:
            logger.info('create elbv2 listener failed')
            return False
        logger.info('create elbv2 listener succeeded')
        return res

    def delete_listener(self, listener_id):
        logger.info('begin to delete elbv2 listener, id is: %s' % listener_id)
        res = process_request(ELBV2_LISTENER_API % listener_id, 'elbv2', 'delete')
        if res == False:
            logger.error('delete elbv2 listener failed, id is %s' % listener_id)
            return False
        logger.info('delete elbv2 listener succeeded')
        return res

    def create_pool(self, listener_id, listener_info):
        logger.info('begin to create elbv2 pool...')
        algorithm = listener_info['lb_algorithm']
        post_data = {
            'pool': {
                'lb_algorithm': self.lb_algorithm_mappings[algorithm],
                'listener_id': listener_id,
                'protocol': listener_info['backend_protocol']
            }
        }
        if listener_info['session_sticky']:
            cookie_name = 'SOURCE_IP' if listener_info['protocol'] == 'TCP' else 'HTTP_COOKIE'
            post_data['pool']['session_persistence'] = {
                'type': cookie_name
            }
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_POOLS_API, 'elbv2', 'post', post_data)
        if res == False:
            logger.info('create elbv2 pool failed')
            return False
        logger.info('create elbv2 pool succeeded')
        return res

    def delete_pool(self, pool_id):
        logger.info('begin to delete elbv2 pool, id is: %s' % pool_id)
        res = process_request(ELBV2_POOL_API % pool_id, 'elbv2', 'delete')
        return res

    def create_member(self, pool_id, member_subnet_id, backend_port, member_info):
        logger.info('begin to create elbv2 member...')
        post_data = {
            'member': {
                'subnet_id': member_subnet_id,
                'protocol_port': backend_port,
                'address': member_info['server_address']
            }
        }
        if member_info['server_name']:
            post_data['member']['name'] = member_info['server_name']
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_MEMBERS_API % pool_id, 'elbv2', 'post', post_data)
        if res == False:
            logger.info('create elbv2 member failed')
            return False
        logger.info('create elbv2 member succeeded')
        return res

    def delete_member(self, pool_id, member_id):
        logger.info('begin to delete elbv2 member, id is: %s' % member_id)
        res = process_request(ELBV2_MEMBER_API % (pool_id, member_id), 'elbv2', 'delete')
        return res

    def create_healthcheck(self, pool_id, healthcheck_info):
        logger.info('begin to create elbv2 healthcheck...')
        post_data = {
            'healthmonitor': {
                'pool_id': pool_id,
                'max_retries': healthcheck_info['unhealthy_threshold'],
                'timeout': healthcheck_info['healthcheck_timeout'],
                'type': healthcheck_info['healthcheck_protocol'],
                'delay': healthcheck_info['healthcheck_interval'],
                'monitor_port': healthcheck_info['healthcheck_connect_port']
            }
        }
        if healthcheck_info['healthcheck_uri']:
            post_data['healthmonitor']['url_path'] = healthcheck_info['healthcheck_uri']
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_HEALTHMONITORS_API, 'elbv2', 'post', post_data)
        if res == False:
            logger.info('create elbv2 healthcheck failed')
            return False
        logger.info('create elbv2 healthcheck succeeded')
        return res

    def delete_healthcheck(self, _id):
        logger.info('begin to delete elbv2 healthcheck, id is: %s' % _id)
        res = process_request(ELBV2_HEALTHMONITOR_API % _id, 'elbv2', 'delete')
        return res

    def create_certificate(self, cert_info):
        logger.info('begin to create elbv2 certificate...')
        post_data = {
            'name': cert_info['name'],
            'description': cert_info['description'],
            'certificate': cert_info['certificate'],
            'private_key': cert_info['private_key']
        }
        if cert_info['domain'] and cert_info['domain'] != 'null':
            post_data['domain'] = cert_info['domain']
        post_data = json.dumps(post_data)
        res = process_request(ELBV2_CERTIFICATES_API, 'elbv2', 'post', post_data)
        if res == False:
            logger.error('create elbv2 certificate failed')
            return False
        logger.info('create elbv2 certificate succeeded')
        return res
    
    def associate_fip(self, lb_vip, eip):
        pass

    def create_rollback(self, created_resource):
        if not created_resource:
            return
        elif 'listeners' in created_resource:
            for listener_id in created_resource['listeners']:
                if 'pool_id' in created_resource['listeners'][listener_id]:
                    pool_id = created_resource['listeners'][listener_id]['pool_id']
                    if 'members' in created_resource['listeners'][listener_id]:
                        for member_id in created_resource['listeners'][listener_id]['members']:
                            self.delete_member(pool_id, member_id)
                    if 'healthcheck_id' in created_resource['listeners'][listener_id]:
                        healthcheck_id = created_resource['listeners'][listener_id]['healthcheck_id']
                        self.delete_healthcheck(healthcheck_id)
                    self.delete_pool(pool_id)
                self.delete_listener(listener_id)
        self.delete_lb(created_resource['lb'])

    def print_migrate_result(self):
        result_logger.info('\n')
        result_logger.info('show elbv2 create result:')
        result_logger.info('*' * 40 + ' CREATE RESULT' + '*' * 40)
        if self.created_mapping_resource:
            result_logger.info('loadbalancers below successfully created in elbv2:')
            result_logger.info('[WARNING]the vip of loadbalancer in elbv2 will be different with elbv1')
            result_logger.info('-' * 75)
            result_logger.info('| elbv1 loadbalancer id' + ' ' * 12 + 
                               '| elbv2 loadbalancer id' + ' ' * 14 + '  |')
            result_logger.info('-' * 75)
            for elbv1_lb_id in self.created_mapping_resource:
                result_logger.info('| ' + elbv1_lb_id + ' | ' +
                                   self.created_mapping_resource[elbv1_lb_id] + ' |')
                result_logger.info('-' * 75)
        if self.created_failed_lbs:
            result_logger.info('\n')
            result_logger.info('loadbalancers below failed to create in elbv2:')
            result_logger.info('-' * 37)
            result_logger.info('| elbv1 loadbalancer id' + ' ' * 13 + '|')
            result_logger.info('-' * 37)
            for elbv1_lb_id in self.created_failed_lbs:
                result_logger.info('| ' + elbv1_lb_id + '  | ')
                result_logger.info('-' * 37)
        result_logger.info('*' * 45 + ' END ' + '*' * 45)
        
    def migrate(self):
        try:
            certificate_id_mappings = {}
            for certificate in self.certificates:
                res = self.create_certificate(certificate)
                if res:
                    certificate_id_mappings[certificate['id']] = res['id']
                else:
                    sys.exit(1)
            for lb in self.lb_infos:
                subnet_id = self.check_valid_vip(lb['vpc_id'])
                if subnet_id:
                    created_resource = {}
                    try:
                        elbv1_lb_id = lb['id']
                        elbv2_lb = self.create_lb(subnet_id, lb)
                        if not elbv2_lb:
                            logger.error('create elbv2 lb failed, mapping elbv1 lb id is: %s' % elbv1_lb_id)
                            continue
                        elbv2_lb_id = elbv2_lb['loadbalancer']['id']
                        created_resource['lb'] = elbv2_lb_id
                        vip_subnet_id = elbv2_lb['loadbalancer']['vip_subnet_id']
                        if 'listeners' in lb:
                            for listener in lb['listeners']:
                                cert_id = listener['certificate_id']
                                if cert_id:
                                    listener['default_tls_container_ref'] = \
                                        certificate_id_mappings[cert_id]
                                elbv2_listener = self.create_listener(elbv2_lb_id, listener)
                                elbv2_listener_id = elbv2_listener['listener']['id']
                                if 'listeners' in created_resource:
                                    created_resource['listeners'][elbv2_listener_id] = {}
                                else:
                                    created_resource['listeners'] = {elbv2_listener_id: {}}
                                backend_port = listener['backend_port']
                                elbv2_pool = self.create_pool(elbv2_listener_id, listener)
                                elbv2_pool_id = elbv2_pool['pool']['id']
                                created_resource['listeners'][elbv2_listener_id]['pool_id'] = \
                                    elbv2_pool_id
                                if 'members' in listener:
                                    for member in listener['members']:
                                        member_subnet_id = self.get_member_subnet_id(member['server_address'])
                                        elbv2_member = self.create_member(elbv2_pool_id, 
                                            member_subnet_id, backend_port, member)
                                        elbv2_member_id = elbv2_member['member']['id']
                                        if 'members' in created_resource['listeners'][elbv2_listener_id]:
                                            created_resource['listeners'][elbv2_listener_id]['members'].append(elbv2_member_id)
                                        else:
                                            created_resource['listeners'][elbv2_listener_id]['members'] = \
                                                [elbv2_member_id]
                                if 'healthcheck' in listener:
                                    elbv2_healthcheck = self.create_healthcheck(elbv2_pool_id, listener['healthcheck'])
                                    elbv2_healthcheck_id = elbv2_healthcheck['healthmonitor']['id']
                                    created_resource['listeners'][elbv2_listener_id]['healthcheck_id'] = elbv2_healthcheck_id
                        self.created_mapping_resource[elbv1_lb_id] = elbv2_lb_id
                        self.created_lbs[elbv1_lb_id] = elbv2_lb['loadbalancer']
                    except Exception as exc:
                        self.create_rollback(created_resource)
                else:
                    logger.error('Can not find valid subnet to create elbv2 loadbalancer')
                
        except Exception as exc:
            raise exc
        finally:
            self.created_failed_lbs = [id for id in self.lb_ids if id not in self.created_lbs]
            self.print_migrate_result()
            

class delete_elbv1_resource(object):

    def __init__(self, created_lb_mappings, elbv1_lb_infos, elbv1_lb_certs, elbv2_lbs):
        self.created_lb_mappings = created_lb_mappings
        self.lb_infos = elbv1_lb_infos
        self.lb_certs = elbv1_lb_certs
        self.elbv2_lbs = elbv2_lbs
        self.created_lb_mappings = created_lb_mappings
        self.success_deleted_lbs = []
        self.associate_fip_failed_lbs = []

    def delete_elbv1_lb(self, _id, lb_type):
        logger.info('begin to delete elbv1 lb, id is: %s' % _id)
        delete_api = KEEP_EIP_DELETE_API if lb_type == 'External' else LB_API
        res = process_request(delete_api %(TENANT_ID, _id), 'elbv1', 'delete')
        check_job_times = 200
        times = 0
        if res:
            while(times < check_job_times):
                job = process_request(JOB_API % (TENANT_ID, res['job_id']))
                if job['status'] == 'SUCCESS':
                    return True
                elif job['status'] == 'FAIL':
                    break
                times += 1
                time.sleep(2)
        result_logger.info('delete elbv1 loadbalancer %s failed' % _id)
        return False

    def delete_elbv1_member(self, listener_id, member_id):
        logger.info('begin to delete elbv1 member, id is: %s' % member_id)
        post_data = {
            "removeMember": [
                {
                    "id": member_id
                }
            ]
        }
        post_data = json.dumps(post_data)
        process_request(MEMBER_API % (TENANT_ID, listener_id), 'elbv1', 'post', post_data)
    
    def print_delete_result(self):
        failed_deleted_lbs = [id for id in self.created_lb_mappings if id not in self.success_deleted_lbs]
        result_logger.info('\n')
        result_logger.info('show elbv1 delete result:')
        result_logger.info('*' * 40 + ' DELETE RESULT' + '*' * 40)
        if self.success_deleted_lbs:
            result_logger.info('loadbalancers below are successfully deleted in elbv1:')
            result_logger.info('-' * 37)
            result_logger.info('| elbv1 loadbalancer id' + ' ' * 13 + '|')
            result_logger.info('-' * 37)
            for id in self.success_deleted_lbs:
                result_logger.info('| ' + id + '  |')
                result_logger.info('-' * 37)
        if failed_deleted_lbs:
            result_logger.info('\n')
            result_logger.info('loadbalancers below are delete failed in elbv1:')
            result_logger.info('-' * 37)
            result_logger.info('| elbv1 loadbalancer id' + ' ' * 13 + '|')
            result_logger.info('-' * 37)
            for id in failed_deleted_lbs:
                result_logger.info('| ' + id + '  |')
                result_logger.info('-' * 37)
        for item in self.associate_fip_failed_lbs:
            result_logger.info('please associate %s to loadbalancer %s manually' % (item['ip'], item['id']))
        result_logger.info('*' * 45 + ' END ' + '*' * 45)
    
    def associate_fip_to_elbv2(self, ip, elbv2_id, elbv2_info):
        logger.info('begin to associate ip %s to elbv2 lb, lb id is %s' %(ip, elbv2_id))
        fips = process_request(FIPS_API, 'fip')
        for fip in fips['floatingips']:
            if fip['floating_ip_address'] == ip:
                fip_id = fip['id']
                break
        else:
            logger.info('\n')
            result_logger.error('[ERROR] can not find eip: %s' % ip)
            result_logger.error('[WARNING] please associate fip to elbv2 lb manually, lb id is %s' % elbv2_id)
            return False
        fip_port_id = elbv2_info['vip_port_id']
        post_data = {
            'floatingip':{
                'port_id': fip_port_id
            }
        }
        post_data = json.dumps(post_data)
        result = process_request(FIP_API % fip_id, 'fip', 'put', post_data)
        if not result:
            result_logger.error('associate fip %s to elbv2 lb failed, lb id is %s' % (ip, elbv2_id))
            item = {
                'id': elbv2_id,
                'ip': ip
            }
            self.associate_fip_failed_lbs.append(item)

    def delete_v1_resource(self):
        logger.warn('begin to delete elbv1 resource')
        for lb in self.lb_infos:
            if lb['id'] in self.created_lb_mappings:
                try:
                    if lb['type'] == 'External':
                        if 'listeners' in lb:
                            for listener in lb['listeners']:
                                if 'members' in listener:
                                    for member in listener['members']:
                                        self.delete_elbv1_member(listener['id'], member['id'])
                    if self.delete_elbv1_lb(lb['id'], lb['type']):
                        self.success_deleted_lbs.append(lb['id'])
                    if lb['type'] == 'External':
                        ip = lb['vip_address']
                        elbv2_id = self.created_lb_mappings[lb['id']]
                        print('\n')
                        input_str = raw_input("Do you want to associate %s to elbv2? lb id is %s [yes|no]" % (ip, elbv2_id))
                        while 1:
                            if input_str not in ['yes', 'no']:
                                print("Wrong Input!!! You can only input 'yes' or 'no'")
                            else:
                                break
                        if input_str == 'yes':
                            self.associate_fip_to_elbv2(ip, elbv2_id, self.elbv2_lbs[lb['id']])
                except:
                    logger.warn('delete elbv1 loadbalancer failed, lb id is %s:' % lb['id'])
        self.print_delete_result()

    
def main():
    parser = argparse.ArgumentParser(
        description='check elbv1 resrouce is ready to migrated or not, '
                    'and migrate resource to elbv2 if necessary')
    parser.add_argument("--lb", default='all', help='the id of '
                        'loadbalancer, different lb id must be seperated by space')
    parser.add_argument("--dry-run", action='store_true', help='Validate that the elbv1 '
                        'LoadBalancer is compatible with elbv2, but do not perform create '
                        'elbv2 or delete elbv1 operations ')
    signal.signal(signal.SIGINT, user_exit)
    signal.signal(signal.SIGTERM, user_exit)
    args = parser.parse_args()
    dry_run = args.dry_run
    lb = args.lb
    logger.info('Begin to run the script...')
    lb_check = check(lb)
    if not lb_check.print_result():
        sys.exit(1)
    if not dry_run:
        if lb_check.invalid_lbs:
            sys.exit()
        if not lb_check.valid_lbs:
            logger.warn('you have no available loadbalancer to migrate to elbv2')
            sys.exit()
        while(1):
            print('\n')
            input_str = raw_input("Are you sure to create loadbalancers in elbv2? relevent elbv1 loadbalancer "
                            "id is in " + str(lb_check.valid_lbs) + ' [yes|no]')
            if input_str not in ['yes', 'no']:
                print("Wrong Input!!! You can only input 'yes' or 'no'")
            else:
                break
        if input_str == 'yes':
            lb_num = len(lb_check.valid_lbs)
            listener_num = lb_check.listener_num
            lb_migrate = migration(lb_check.valid_lbs, lb_check.lbs, lb_check.certificates)
            lb_migrate.migrate()
        else:
            sys.exit()
        if lb_migrate.created_lbs:
            while(1):
                print('\n')
                input_str = raw_input("Are you sure to delete loadbalancers in elbv1? loadbalancer "
                                "id is in " + str(lb_migrate.created_lbs.keys()) + ' [yes|no]')
                if input_str not in ['yes', 'no']:
                    print("Wrong Input!!! You can only input 'yes' or 'no'")
                else:
                    break
            if input_str == 'yes':
                lb_delete = delete_elbv1_resource(lb_migrate.created_mapping_resource, 
                                                  lb_check.lbs, lb_check.certificates, lb_migrate.created_lbs)
                lb_delete.delete_v1_resource()
            else:
                sys.exit()


if __name__ == '__main__':
    main()
